/*********************************\
 *  vector reduction using CUDA  *
 *  Uses a 1D grid layout        *
\*********************************/
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <device_launch_parameters.h> // cuda keywords
#include <cuda_runtime.h> // cuda runtime functions
#include <cuda.h>

#include "utility_environment.h"  // some helper functions, the names of which begin with "ue_"


// to make this constant available to both host and device
const unsigned int threads_per_block = 256;						// any 2^k in [1,1024]


/************************************************************************\
 *  kernel to compute sum, max, and argmax of a vector of length n      *
 *                                                                      *
 *  each block is reduced in parallel, but the intermediate results of  *
 *  all blocks must be combined on the host                             *
\************************************************************************/
__global__
void reduceMaxSumGPU( const float *src, float *dstSum, float *dstMax, unsigned int *dstArg, unsigned int n )
{
	// ** TODO **

}
/*****************************************************************\
 *  compute the next highest power of 2 of 32-bit integer value  *
\*****************************************************************/
unsigned int roundUpPowerOfTwo(unsigned int v)
{
	v--;             // subtract one so that we round powers of two onto themselves
	v |= v >> 1;     // "smear" the leading 1 bit across all lower bits to the right
	v |= v >> 2;     //
	v |= v >> 4;     // fun fact:
	v |= v >> 8;     // this algorithm is a close relative of the reduction we do on the GPU
	v |= v >> 16;
	v++;

	return v;
}


/***********\
 *  setup  *
\***********/
int main( int argc, char **argv )
{
	//////////////////////////////////////////////////
	// read vector length as command line parameter //
	//////////////////////////////////////////////////
	unsigned int vec_len = 8192;									// just a default
	if ( argc > 1 )
	{
		// interpret first argument as length
		vec_len = atoi( argv[1] );
	}
	// round up to a power of two
	vec_len = roundUpPowerOfTwo(vec_len);
	printf( "compute sum, max, and argmax of a vector of length %d\n", vec_len );

	///////////////////////////////////////////////////////////////////////////////
	// set layout of blocks and threads; we make a 1D threads/blocks layout here //
	///////////////////////////////////////////////////////////////////////////////
	unsigned int num_blocks = (vec_len + threads_per_block - 1) / threads_per_block;
	// this funny expression yields ceil( max_blocks_per_grid / threads_per_block ) in integer arithmetic
	printf( "Threads per block = %d\n", threads_per_block );
	printf( "Number of blocks  = %d\n", num_blocks );

	// check maximally allowed layout sizes
	unsigned int max_n_threads_per_block, max_n_threads_x, max_n_blocks_x;
	ue_getMaxProperties( & max_n_threads_per_block, & max_n_threads_x, & max_n_blocks_x );
	if ( threads_per_block > max_n_threads_per_block || threads_per_block > max_n_threads_x )
	{
		fprintf(stderr,"Too many threads per block! Max = %d / %d\n", max_n_threads_per_block, max_n_threads_x );
		exit(2);
	}
	if ( num_blocks > max_n_blocks_x )
	{
		fprintf(stderr,"Too many blocks in grid row! Max = %d\n", max_n_blocks_x );
		exit(2);
	}

	///////////////////////////////////////////////////////////
	// fill vector with random numbers, setup device buffers //
	///////////////////////////////////////////////////////////
	srand48( unsigned(time(NULL)) );
	size_t size = vec_len * sizeof(float);
	float* h_vec = new float[vec_len];
	float* h_sum = new float[num_blocks];
	float* h_max = new float[num_blocks];
	unsigned int* h_arg = new unsigned int[num_blocks];
	if ( h_vec == 0 || h_sum == 0 || h_max == 0 || h_arg == 0 )
    {
        fprintf(stderr, "Failed to allocate host vectors!\n");
		exit(1);
	}
	
    for ( unsigned int i = 0; i < vec_len; i++ )
    {
		h_vec[i] = 10.0f * float(drand48() - 0.5f);    // random numbers in range [-5.0 .. 5.0]
    }
	
	float *d_vec, *d_sum, *d_max;
	unsigned int *d_arg;
	cudaMalloc( (void **) &d_vec, size );
	cudaMalloc( (void **) &d_sum, num_blocks * sizeof(float) );
	cudaMalloc( (void **) &d_max, num_blocks * sizeof(float) );
	cudaMalloc( (void **) &d_arg, num_blocks * sizeof(int) );
	handleCUDAerror(__LINE__);

	// transfer the initialized vector to the device
	cudaMemcpy( d_vec, h_vec, size, cudaMemcpyHostToDevice );
	handleCUDAerror(__LINE__);

	////////////////////////////////////////
	// launch the vector reduction kernel //
	////////////////////////////////////////
	reduceMaxSumGPU<<<num_blocks,threads_per_block>>>( d_vec, d_sum, d_max, d_arg, vec_len );
	handleCUDAerror(__LINE__);

	// copy results back to the host (does cudaDeviceSynchronize implicitely)
	cudaMemcpy( h_sum, d_sum, num_blocks * sizeof(float), cudaMemcpyDeviceToHost );
	cudaMemcpy( h_max, d_max, num_blocks * sizeof(float), cudaMemcpyDeviceToHost );
	cudaMemcpy( h_arg, d_arg, num_blocks * sizeof(unsigned int), cudaMemcpyDeviceToHost );
	handleCUDAerror(__LINE__);

	// collect all block results on the host
	float sumGPU = 0.0f;
	float maxGPU = h_max[0];
	unsigned int argmaxGPU = h_arg[0];
	for (unsigned int i = 0; i < num_blocks; i++)
	{
		sumGPU += h_sum[i];
		if (h_max[i] > maxGPU)
		{
			maxGPU = h_max[i];
			argmaxGPU = h_arg[i];
		}
	}

	//////////////////////////////////////////
	// compute reference result on the host //
	//////////////////////////////////////////
	float sumCPU = 0.0f;
	float maxCPU = h_vec[0];
	unsigned int argmaxCPU = 0;
	for (unsigned int i = 0; i < vec_len; i++)
	{
		sumCPU += h_vec[i];
		if (h_vec[i] > maxCPU)
		{
			maxCPU = h_vec[i];
			argmaxCPU = i;
		}
	}

	// display both sets of results
	printf("CPU: sum: %3.5f\t max: %3.5f\t argmax: %3d\nGPU: sum: %3.5f\t max: %3.5f\t argmax: %3d\n", sumCPU, maxCPU, argmaxCPU, sumGPU, maxGPU, argmaxGPU);


	//////////////
	// clean up //
	//////////////
	cudaFree( d_vec );
	cudaFree( d_sum );
	cudaFree( d_max );
	cudaFree( d_arg );
	handleCUDAerror(__LINE__);

	free( h_vec );
	free( h_sum );
	free( h_max );
	free( h_arg );

	return 0;
}

